with Strings, Globals, ScalVectors.Eigenv, ScalVectors.Eigenv.RNum;
use Strings, Globals;

pragma Elaborate_All (Strings,Globals,ScalVectors.Eigenv.RNum);

package body TSpherics.Fix is

  use Lin,Lin_Contr;

  procedure LinMap(LF: in out Spheric) is
  begin
    case Map is
      when 0 => NegInvLap(LF);
      when 1 => InvPlate(LF);
      when 2 => InvWave(LF);
      when 3 => RotWave(LF);
      when 4 => InvQP(LF);
      when others => raise Not_Implemented;
    end case;
  end LinMap;

  procedure GMap(F: in Spheric; GF: in out Spheric) is
  begin
    Prod(F,F,GF);
    Mult(F,GF);
    LinMap(GF);
  end GMap;

  procedure DGMap(F,FF3,H: in Spheric; DGH: in out Spheric) is
    pragma Unreferenced(F);
    --- FF3 = 3*F*F
  begin
     Prod(H,FF3,DGH);
     LinMap(DGH);
  end DGMap;

  procedure DContrNorm(M: in SMatrix; F: in Spheric; PackCSize,PackSSize: in Natural; R: out Radius) is
    Q: constant TSModes := Make(PackCSize,PackSSize,False);  --- Err modes
    FF3: Spheric;
    procedure DG(H: in Spheric; DGH: in out Spheric) is begin DGMap(F,FF3,H,DGH); end DG;
    procedure DContract is new DContr (DMap => DG);
    procedure DC(H: in Spheric; DCH: in out Spheric) is begin DContract(QC,M,H,DCH); end DC;
    procedure DCNorm is new Op_Norm (LinOp => DC);
  begin
    if PackSSize <= Mat_PackSSize then raise Not_Implemented; end if;
    if PackCSize < Mat_PackCSize then raise Not_Implemented; end if;
    if PackCSize = Mat_PackCSize and then PackCSize/=0 then raise Not_Implemented; end if;
    Prod(F,F,FF3);
    Mult(SThree,FF3);
    if Verbosity>1 then
      Show1("DContrNorm: Q'Last = ",Q'Last);
    end if;
    DCNorm(Q,R);
  end DContrNorm;

  procedure ContrFix(M: in SMatrix; PackCSize,PackSSize: in Natural; F: in out Spheric) is
    Tiny: constant Radius := Flt'Scaling(One,Flt'Machine_Emin);
    KMax: constant Radius := Flt(7)/Flt(8);
    E,K: Radius;
    ANorm,R,Eps,Del: Scalar;
    F0: Spheric;
  begin
    Show0("ContrFix: ",False);
    Copy(F,F0);
--    Center(F0);
    GMap(F0,F);
    Sub(F0,F);
    E := MaxNorm(F);
    Assign(E,Eps);
    OpNorm(QC,M,ANorm);
    Add(SOne,ANorm);
    Show2("Eps,ANorm ",E,Sup(ANorm));
    ---
    Del := Eps/(SOne-Scal(KMax))+Scal(Tiny);
    R := ANorm*Del;
    Copy(F0,F);
    AddBall(MaxNorm(R),F);
    DContrNorm(M,F,PackCSize,PackSSize,K);
    Show1("ContrFix: K ",K);
    if K>KMax then raise Sorry with "K too large"; end if;
    ---
    Del := Eps/(SOne-Scal(K))+Scal(Tiny);
    R := ANorm*Del;
    Show2("ContrFix: Del,R: ",Sup(Del),Sup(R));
    Copy(F0,F);
    AddBall(MaxNorm(R),F);
  end ContrFix;

  --------------------------------------------
  --------------------------------------------

  procedure DerMatrix(F: in Spheric; M: in out SMatrix) is
    --- for numeric mode only
    FF3: Spheric;
    procedure DG(H: in Spheric; DGH: in out Spheric) is begin DGMap(F,FF3,H,DGH); end DG;
    procedure Enclose is new Enclosure (LinOp => DG);
  begin
    if Not_Trunc then raise Sorry; end if;
    Prod(F,F,FF3);
    Mult(SThree,FF3);
    Enclose(QC,M);
  end DerMatrix;

  procedure ContrMatrix(F: in Spheric; M: in out SMatrix) is
  begin
    if Verbosity>1 then Show0("running ContrMatrix"); end if;
    DerMatrix(F,M);
    Contr_Matrix(M);
  end ContrMatrix;

  procedure ApproxFix(F: in out Spheric; M: in out SMatrix; Eps: in Radius; MinErr: in out Radius) is
    --- find approximate fixed point
    procedure Contract0 is new ContrTmp0 (Map => GMap);
    Iter: constant Integer := 1024;
    MaxCredit: constant Integer := 5;
    Credit: Integer := MaxCredit;
    Err: Radius := MinErr;
    ErrMat: constant Radius := 1.0e-4;
    G,Tmp1,Tmp2: Spheric;
  begin
    if Not_Trunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running Approx_Fix"); end if;
    Copy(F,G);
    for I in 1 .. Iter loop
      if Err>ErrMat then ContrMatrix(G,M); end if;
      Contract0(QC,M,G,Tmp1,Tmp2,Err);
      if Verbosity>1 then Show1("ApproxFix: step " & Strng(I) & ", Err ",Err); end if;
      if Err<MinErr then
        Copy(G,F);
        MinErr := Err;
        if (Err<Eps) then exit; end if;
        Credit := MaxCredit;
      else
        exit when Credit=0;
        Credit := Credit-1;
      end if;
    end loop;
    --    ContrMatrix(F,M);
  end ApproxFix;

  procedure ApproxFix(F: in out Spheric; M: in out SMatrix; Eps: in Radius := Zero) is
    MinErr: Radius := 1.0E+9;
  begin
    ApproxFix(F,M,Eps,MinErr);
    Show1("ApproxFix: Err ",MinErr);
  end ApproxFix;

  procedure IterContr(F: in out Spheric; M: in out SMatrix) is
    --- improve approximate fixed point
    procedure Contract0 is new ContrTmp0 (Map => GMap);
    MaxCredit: constant Integer := 16;
    Credit: Integer := MaxCredit;
    Err,MinErr: Radius := One;
    G,Tmp1,Tmp2: Spheric;
  begin
    if Not_Trunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running IterContr"); end if;
    Copy(F,G);
    --ContrMatrix(G,M);
    while Credit>0 loop
      Contract0(QC,M,G,Tmp1,Tmp2,Err);
      Show1("IterContr: Err ",Err);
      if Err<MinErr then
        Copy(G,F);
        MinErr := Err;
        Credit := MaxCredit;
        --ContrMatrix(G,M);
      else
        Credit := Credit-1;
      end if;
    end loop;
    Show1("IterContr: Err ",MinErr);
    --ContrMatrix(F,M);
  end IterContr;

  procedure EigenDer(Name: in String; F: in Spheric) is
    A,Vr,Vi: QC_Matrix;
    Wr,Wi: QC_Vector;
    FHat: Spheric;
    package SVE is new SSV.Eigenv;
    package SVEN is new SVE.RNum;
    use SV,SSV,SVEN;

    procedure AddProd(V: in QC_Vector) is
    begin
      for I in V'Range loop
        AddProd(QC(I),V(I),FHat,False);
      end loop;
    end AddProd;

  begin
    DerMatrix(F,A);
    Eigen(A,Wr,Wi,Vr,Vi);
    SetZero(FHat);
    for I in 1..IMin(50,Wr'Last) loop
      if Approx(Wr(I))>0.5 then
        Show2("ev:",Wr(I),Wi(I));
      end if;
      if abs(Approx(Wr(I))-One)<0.01 then
        AddProd(GetColumn(I,Vr));
      end if;
    end loop;
    --    Write(Name,FHat);
  end EigenDer;

end TSpherics.Fix;
