with Globals, Ints;
use Globals, Ints;

pragma Elaborate_All (Globals,Ints);

package body TSpherics_Aux is

  ------------------------------
  --- periodic wave equation ---
  ------------------------------

  procedure Wave_Del(L: in Positive; Del: out Radius; Tmp,Tmp1,Tmp2: in out Scalar) is
    --- lower bound on Delta_L for a single L
    --- do not use with Trunc
    NMin,NMax: Natural;
    Del1,Del2: Radius;
    B: Scalar renames Tmp1;
    S: Scalar renames Tmp2;
  begin
    Assign(L*(L+1),Tmp);
    Sqrt(Tmp,B);                          --- B=Sqrt{L*(L+1)}
    Prod(SInvAlpha,B,Tmp);                --- B/Alpha
    NMin := Integer(Flt'Floor(Inf(Tmp)));
    NMax := Integer(Flt'Ceiling(Sup(Tmp)));
    if (NMax-NMin)/=1 then raise Sorry; end if;
    Prod(Flt(NMin),SAlpha,Tmp);
    Diff(B,Tmp,S);             --- B-Alpha*NMin should be pos
    Del1 := Inf(S);
    Prod(Flt(NMax),SAlpha,Tmp);
    Diff(Tmp,B,S);             --- Alpha*NMax-B should be pos
    Del2 := Inf(S);
    Del := RMin(Del1,Del2);    --- Del=Min(B-Alpha*NMin,Alpha*NMax-B)
  end Wave_Del;

  procedure Wave_Eps(L: in Positive; Eps,Tmp: in out Scalar) is
    --- do not use with Trunc
    Den1: constant Positive := 2*L+1;
    Den2: constant Positive := 4*(Den1*Den1-1); --- 4*[(2*L+1)^2-1]
  begin
    Assign(Den2,Eps);
    Inv(Eps,Tmp);
    Add(1,Tmp);                -- Tmp=1+1/Den2
    Quot(Tmp,Flt(4*Den1),Eps); -- Eps=(1+1/Den2)/(4*Den1)
  end Wave_Eps;

  procedure Make_Wave_Del(Del: out RadVec) is
    --- do not use with Trunc
    --- assuming Del'First=0
    SizeE: constant Integer := Del'Last;
    Tmp,Tmp1,Tmp2: Scalar;
    Eps: Scalar renames Tmp1;
    HiDel: Scalar renames Tmp2;
  begin
    Del(0) := Inf(SAlpha);               --- assumes N>0 if L=0 !!!
    for L in 1 .. SizeE-1 loop
      Wave_Del(L,Del(L),Tmp,Tmp1,Tmp2);
    end loop;
    Wave_Eps(SizeE,Eps,Tmp);
    Prod(Flt(AlphaQ),Eps,Tmp);           --- Tmp=Q*Eps
    Assign(Half,Tmp1);
    Diff(Tmp1,Tmp,HiDel);                --- Del=Half-Q*Eps
    Del(SizeE) := Inf(HiDel);            --- lower bound on all Delta_L with L>=SizeE
  end Make_Wave_Del;

  function Make_Wave_Del(SizeE: Positive) return RadVec is
    Del: RadVec(0 .. SizeE);
  begin
    Make_Wave_Del(Del);
    return Del;
  end Make_Wave_Del;

  -------------------------------
  --- periodic plate equation ---
  -------------------------------

    procedure Plate_Del(L: in Positive; Del: out Radius) is
    --- lower bound on Delta_L for a single L
    --- use only if FreqPar<0 or Trunc
  begin
    if Trunc then
      Del := Zero;
    else
      declare
        AlphaP2: constant Integer := 2*AlphaP;
        U: Integer;
      begin
        U := (AlphaQ*L*(L+1)-AlphaP) mod AlphaP2;
        if U>AlphaP then U := AlphaP2-U; end if;
        Del := -(Flt(-U)/Flt(AlphaQ));               --- U/Q
      end;
    end if;
  end Plate_Del;

  -------------------------------------
  --- quasi-periodic plate equation ---
  -------------------------------------

  procedure Make_Beta(I: in Integer; J: in Positive; Beta: in out Scalar) is
    --- Beta=|I+Theta|, Theta=[Sqrt(J*J+4)-J]/2 = 2/[Sqrt(J*J+4)+J]
    Tmp: Scalar;
  begin
    Assign(J*J+4,Tmp);
    Sqrt(Tmp,Beta);         -- Sqrt(J*J+4)
    Assign(J,Tmp);
    Add(Beta,Tmp);          -- Sqrt(J*J+4)+J
    Inv(Tmp,Beta);
    Mult(Two,Beta,Tmp);     -- Theta = 2/[Sqrt(J*J+4)+J]
    if I/=0 then
      Assign(I,Tmp);
      Add(Tmp,Beta);
      if I<0 then Neg(Beta); end if;
    end if;
  end Make_Beta;

  function Make_Beta(I: Integer; J: Positive) return Scalar is
    Beta: Scalar;
  begin
    Make_Beta(I,J,Beta);
    return Beta;
  end Make_Beta;

  function Inf(Q: Rational) return Flt is
    NegNum: constant LInt := -LNum(Q);
    Den: constant LPos := LDen(Q);
  begin
    return -(Flt(NegNum)/Flt(Den));
  end Inf;

  function Dio_Const(J: Positive; N: Natural) return Radius is
    --- Beta=[0;j,j,j,...]
    --- lower bound for denominators >= N
    QKMinus: Integer;
    QK: Integer := 0;               --- Q(-1)
    QKPlus: Integer := 1;           --- Q( 0)
    BetaK,BetaKPlus,B: Rational;
  begin
    loop                            --- starting with K=0
      QKMinus := QK;                --- Q(K-1)
      QK := QKPlus;                 --- Q(K)
      QKPlus := J*QK+QKMinus;       --- Q(K+1) = J*Q(K)+Q(K-1)
      exit when QKPlus >= N;
    end loop;
    BetaK := IRat(QKMinus,QK);    --- Beta_k = P_k/Q_k
    BetaKPlus := IRat(QK,QKPlus); --- Beta_{k+1} = P_{k+1}/Q_{k+1}
    if BetaK<BetaKPlus then
      B := Inv(2*BetaKPlus+1);
    else
      B := Inv(2*BetaK+1);
    end if;                       --- B = 1/[1+2*Max(Beta_k,Beta_{k+1})]
    B := N*B/QK;
    return Inf(B);
  end Dio_Const;

  procedure QP_HiSigDel(Beta: in Scalar; L,M: in Positive; N: in out Positive; Sig,Del: out Radius) is
    --- upper bounds on |Sigma_n| and Delta_n for n>=N
    --- input N is lowest acceptable value
    QC: constant Rational := 1/4;      --- denominator has to be a power of 2
    RC: constant Radius := Inf(QC);
    N0: Positive;
    QTmp: Rational;
    Eps: Radius;
    S,Tmp: Scalar;
  begin
    if not (QC<Alpha) then raise Sorry; end if; -- decrease RC
    Assign(Flt(L)+Half,S);
    Prod(S,S,Tmp);
    Quot(Tmp,RC,S);
    N0 := Integer(Flt'Ceiling(Sup(S)));         -- N0>=(L+1/2)^2/C
    if N0>N then N := N0; end if;               -- N>=N0

    Prod(Beta,Beta,Tmp);
    Prod(Flt(M*M),Tmp,S);                       -- S = [Beta*M]^2
    QTmp := (N*N)*Alpha2-IRat(Sqr(L*(L+1)));    -- [Alpha*N]^2-[L(L+1)]^2
    Assign(QTmp,Tmp);
    Add(Tmp,S);                                 -- S = [Beta*M]^2+[Alpha*N]^2-[L(L+1)]^2
    Sig := Inf(S);                              -- lower bound on |sigma_n|

    QTmp := (2*Alpha)/(Alpha2-QC*QC);
    QTmp := (QTmp*QTmp)*(QC/N);
    Assign(QTmp,S);
    Sqrt(S,Tmp);
    Prod(Beta,Tmp,S);                    -- S = Beta*Sqrt(C/N)*2*Alpha/(Alpha^2-C^2)
    Eps := Sup(S);                       -- upper bound on eps_n
    if Eps>=One then raise Sorry; end if;

    Sig := One/Sig;
    Sig := (-Sig)/(Eps*Eps-One);         -- upper bound on |Sigma_n|
    Del := Eps*Sig;                      -- upper bound on Delta_n
  end QP_HiSigDel;

  function SupAbs(S: Scalar) return Radius is
    --- not too efficient
    InfS: constant Flt := Inf(S);
    SupS: constant Flt := Sup(S);
  begin
    return RMax(abs(InfS),abs(SupS));
  end SupAbs;

  function InfAbs(S: Scalar) return Radius is
    --- not too efficient
    InfS: constant Flt := Inf(S);
    SupS: constant Flt := Sup(S);
  begin
    if InfS<=Zero and then SupS>=Zero then return Zero; end if;
    return RMin(abs(InfS),abs(SupS));
  end InfAbs;

  procedure QP_SigDel(Beta: in Scalar; L,M: in Positive; Sig,Del: out RadVec) is
    --- upper bounds on |Sigma_n| and Delta_n for n>=N and N in range of Sig and Del
    EDeg: constant Positive := Imax(Sig'Last,Del'Last);
    SqrLL1: constant Integer := Sqr(L*(L+1));
    N0: Positive := EDeg+1;
    QTmp: Rational;
    MaxSig,MaxDel,Num,Den: Radius;
    BetaM2,Delta1,SigmaN,DeltaN,S,Tmp: Scalar;
  begin
    Prod(Flt(M),Beta,Tmp);
    Prod(Tmp,Tmp,BetaM2);                   -- BetaM2 = [Beta*M]^2
    QTmp := (2*M)*Alpha;
    Assign(QTmp,Tmp);
    Prod(Beta,Tmp,Delta1);                  -- delta_1 = 2*Alpha*Beta*M

    QP_HiSigDel(Beta,L,M,N0,MaxSig,MaxDel); -- upper bounds for N>=N0
    for N in reverse 0 .. N0-1 loop
      QTmp := (N*N)*Alpha2-SqrLL1;          -- [Alpha*N]^2-[L(L+1)]^2
      Assign(QTmp,SigmaN);
      Add(BetaM2,SigmaN);                   -- |sigma_N| = [Beta*M]^2+[Alpha*N]^2-[L(L+1)]^2
      Prod(Flt(N),Delta1,DeltaN);           -- delta_N = 2*Alpha*Beta*M*N

      Prod(DeltaN,DeltaN,Tmp);
      Prod(SigmaN,SigmaN,S);
      Sub(Tmp,S);                           -- S = sigma_N^2-delta_N^2
      Den := InfAbs(S);
      Num := SupAbs(SigmaN);                -- |Sigma_N| = Num/Den (upper bound)
      MaxSig := RMax(Num/Den,MaxSig);       -- upper bound on |Sigma_n| for n>=N
      Num := SupAbs(DeltaN);                -- Delta_N = Num/Den (upper bound)
      MaxDel := RMax(Num/Den,MaxDel);       -- upper bound on Delta_n for n>=N
      if N<=EDeg then
        Sig(N) := MaxSig;
        Del(N) := MaxDel;
      end if;
    end loop;
  end QP_SigDel;

  procedure QP_Lam(I: in Integer; J,NMin: in Positive; Lam: out RadVec) is
    --- lower bound on Lambda_L
    SizeE: constant Positive := Lam'Last;
    Q2: constant Flt := Flt(AlphaQ*AlphaQ);
    SmallBeta: Boolean := True;
    SupBeta: Radius;
    Gamma,Beta,BetaPrime,S,Tmp: Scalar;
  begin
    Assign(Dio_Const(J,NMin),Gamma); -- Gamma
    Make_Beta(I,J,Beta);             -- Beta
    SupBeta := Sup(Beta);
    Lam(0) := Zero;
    for L in 1 .. SizeE loop
      Quot(Gamma,Q2,S);          -- S = Gamma*Q^{-2}
      if SmallBeta then
        Quot(S,Flt(NMin*L),Tmp); -- Gamma/(Q^2*NMin*L)
        if Sup(Tmp)>SupBeta then
          Copy(Tmp,BetaPrime);
        else
          Copy(Beta,BetaPrime);
          SmallBeta := False;
        end if;
      end if;
      Assign(L+1,Tmp);
      Sub(BetaPrime,Tmp);        -- (L+1)-Beta'
      Mult(Tmp,S);               -- Gamma*Q^{-2}[(L+1)-Beta']
      Lam(L) := Two*Inf(S);
    end loop;
  end QP_Lam;

end TSpherics_Aux;
